@@ -15,12 +15,15 @@ use {
         feature_set,
         hash::Hash,
         instruction::{AccountMeta, Instruction, InstructionError},
-        keyed_account::{from_keyed_account, get_signers, keyed_account_at_index, KeyedAccount},
+        keyed_account::{
+            check_sysvar_keyed_account, from_keyed_account, get_signers, keyed_account_at_index,
+            KeyedAccount,
+        },
         process_instruction::{get_sysvar, InvokeContext},
         program_utils::limited_deserialize,
         pubkey::Pubkey,
         system_instruction,
-        sysvar::{self, clock::Clock, slot_hashes::SlotHashes},
+        sysvar::{self, clock::Clock, rent::Rent, slot_hashes::SlotHashes, Sysvar},
     },
     std::collections::HashSet,
     thiserror::Error,
@@ -299,9 +302,8 @@ pub fn withdraw(
 
 fn verify_rent_exemption(
     keyed_account: &KeyedAccount,
-    rent_sysvar_account: &KeyedAccount,
+    rent: &Rent,
 ) -> Result<(), InstructionError> {
-    let rent: sysvar::rent::Rent = from_keyed_account(rent_sysvar_account)?;
     if !rent.is_exempt(keyed_account.lamports()?, keyed_account.data_len()?) {
         Err(InstructionError::InsufficientFunds)
     } else {
@@ -309,6 +311,19 @@ fn verify_rent_exemption(
     }
 }
 
+/// This method facilitates a transition from fetching sysvars from keyed
+/// accounts to fetching from the sysvar cache without breaking consensus. In
+/// order to keep consistent behavior, it continues to enforce the same checks
+/// as `solana_sdk::keyed_account::from_keyed_account` despite dynamically
+/// loading them instead of deserializing from account data.
+fn get_sysvar_with_keyed_account_check<S: Sysvar>(
+    keyed_account: &KeyedAccount,
+    invoke_context: &dyn InvokeContext,
+) -> Result<S, InstructionError> {
+    check_sysvar_keyed_account::<S>(keyed_account)?;
+    get_sysvar(invoke_context, keyed_account.unsigned_key())
+}
+
 pub fn process_instruction(
     _program_id: &Pubkey,
     data: &[u8],
@@ -331,21 +346,24 @@ pub fn process_instruction(
 
     match limited_deserialize(data)? {
         VoteInstruction::InitializeAccount(vote_init) => {
-            verify_rent_exemption(me, keyed_account_at_index(keyed_accounts, 1)?)?;
-            vote_state::initialize_account(
-                me,
-                &vote_init,
-                &signers,
-                &from_keyed_account::<Clock>(keyed_account_at_index(keyed_accounts, 2)?)?,
-            )
+            let rent: Rent = get_sysvar_with_keyed_account_check(
+                keyed_account_at_index(keyed_accounts, 1)?,
+                invoke_context,
+            )?;
+            verify_rent_exemption(me, &rent)?;
+            let clock: Clock = get_sysvar_with_keyed_account_check(
+                keyed_account_at_index(keyed_accounts, 2)?,
+                invoke_context,
+            )?;
+            vote_state::initialize_account(me, &vote_init, &signers, &clock)
+        }
+        VoteInstruction::Authorize(voter_pubkey, vote_authorize) => {
+            let clock: Clock = get_sysvar_with_keyed_account_check(
+                keyed_account_at_index(keyed_accounts, 1)?,
+                invoke_context,
+            )?;
+            vote_state::authorize(me, &voter_pubkey, vote_authorize, &signers, &clock)
         }
-        VoteInstruction::Authorize(voter_pubkey, vote_authorize) => vote_state::authorize(
-            me,
-            &voter_pubkey,
-            vote_authorize,
-            &signers,
-            &from_keyed_account::<Clock>(keyed_account_at_index(keyed_accounts, 1)?)?,
-        ),
         VoteInstruction::UpdateValidatorIdentity => vote_state::update_validator_identity(
             me,
             keyed_account_at_index(keyed_accounts, 1)?.unsigned_key(),
@@ -356,13 +374,15 @@ pub fn process_instruction(
         }
         VoteInstruction::Vote(vote) | VoteInstruction::VoteSwitch(vote, _) => {
             inc_new_counter_info!("vote-native", 1);
-            vote_state::process_vote(
-                me,
-                &from_keyed_account::<SlotHashes>(keyed_account_at_index(keyed_accounts, 1)?)?,
-                &from_keyed_account::<Clock>(keyed_account_at_index(keyed_accounts, 2)?)?,
-                &vote,
-                &signers,
-            )
+            let slot_hashes: SlotHashes = get_sysvar_with_keyed_account_check(
+                keyed_account_at_index(keyed_accounts, 1)?,
+                invoke_context,
+            )?;
+            let clock: Clock = get_sysvar_with_keyed_account_check(
+                keyed_account_at_index(keyed_accounts, 2)?,
+                invoke_context,
+            )?;
+            vote_state::process_vote(me, &slot_hashes, &clock, &vote, &signers)
         }
         VoteInstruction::Withdraw(lamports) => {
             let to = keyed_account_at_index(keyed_accounts, 1)?;
@@ -402,10 +422,11 @@ mod tests {
         bincode::serialize,
         solana_sdk::{
             account::{self, Account, AccountSharedData},
-            process_instruction::{mock_set_sysvar, MockInvokeContext},
+            process_instruction::MockInvokeContext,
             rent::Rent,
+            sysvar_cache::SysvarCache,
         },
-        std::{cell::RefCell, str::FromStr},
+        std::{borrow::Cow, cell::RefCell, str::FromStr},
     };
 
     fn create_default_account() -> RefCell<AccountSharedData> {
@@ -462,12 +483,20 @@ mod tests {
                 .map(|(meta, account)| KeyedAccount::new(&meta.pubkey, meta.is_signer, account))
                 .collect();
             let mut invoke_context = MockInvokeContext::new(keyed_accounts);
-            mock_set_sysvar(
-                &mut invoke_context,
+            let mut sysvar_cache = SysvarCache::default();
+            sysvar_cache.push_entry(
                 sysvar::rent::id(),
-                sysvar::rent::Rent::default(),
-            )
-            .unwrap();
+                bincode::serialize(&Rent::free()).unwrap(),
+            );
+            sysvar_cache.push_entry(
+                sysvar::clock::id(),
+                bincode::serialize(&Clock::default()).unwrap(),
+            );
+            sysvar_cache.push_entry(
+                sysvar::slot_hashes::id(),
+                bincode::serialize(&SlotHashes::default()).unwrap(),
+            );
+            invoke_context.sysvar_cache = Cow::Owned(sysvar_cache);
             super::process_instruction(&Pubkey::default(), &instruction.data, &mut invoke_context)
         }
     }
@@ -656,7 +685,7 @@ mod tests {
 
     #[test]
     fn test_minimum_balance() {
-        let rent = solana_sdk::rent::Rent::default();
+        let rent = Rent::default();
         let minimum_balance = rent.minimum_balance(VoteState::size_of());
         // golden, may need updating when vote_state grows
         assert!(minimum_balance as f64 / 10f64.powf(9.0) < 0.04)
